@@ -130,7 +130,7 @@ char *pystos(PyObject *pys)
 	"Please note that none of the classes' attributes may be deleted.\n" \
 	"This is not a complete mapping to the libmount C API, nor is it\n" \
 	"attempting to be one.\n" "Iterator functions only allow forward\n" \
-	"iteration for now. Contex.get_{user_,}mflags() differs from the C API\n" \
+	"iteration for now. Context.get_{user_,}mflags() differs from the C API\n" \
 	"and returns the flags directly. Fs.get_tag() differs from the C API\n" \
 	"and returns a (tag, value) tuple. Every attribute is \"filtered\"" \
 	"through appropriate getters/setters, no values are set directly."
@@ -210,7 +210,7 @@ PyMODINIT_FUNC initpylibmount(void)
 
 		pylibmount_debug_mask = 0;
 		if (str)
-			pylibmount_debug_mask = strtoul(str, 0, 0);
+			pylibmount_debug_mask = strtoul(str, NULL, 0);
 
 		pylibmount_debug_mask |= PYMNT_DEBUG_INIT;
 	}
@@ -229,7 +229,9 @@ PyMODINIT_FUNC initpylibmount(void)
 
 	FS_AddModuleObject(m);
 	Table_AddModuleObject(m);
+#ifdef __linux__
 	Context_AddModuleObject(m);
+#endif
 
 	/*
 	 * mount(8) userspace options masks (MNT_MAP_USERSPACE map)
@@ -249,6 +251,13 @@ PyMODINIT_FUNC initpylibmount(void)
 	PyModule_AddIntConstant(m, "MNT_MS_USER", MNT_MS_USER);
 	PyModule_AddIntConstant(m, "MNT_MS_USERS", MNT_MS_USERS);
 	PyModule_AddIntConstant(m, "MNT_MS_XCOMMENT", MNT_MS_XCOMMENT);
+	PyModule_AddIntConstant(m, "MNT_MS_HASH_DEVICE", MNT_MS_HASH_DEVICE);
+	PyModule_AddIntConstant(m, "MNT_MS_ROOT_HASH", MNT_MS_ROOT_HASH);
+	PyModule_AddIntConstant(m, "MNT_MS_HASH_OFFSET", MNT_MS_HASH_OFFSET);
+	PyModule_AddIntConstant(m, "MNT_MS_ROOT_HASH_FILE", MNT_MS_ROOT_HASH_FILE);
+	PyModule_AddIntConstant(m, "MNT_MS_FEC_DEVICE", MNT_MS_FEC_DEVICE);
+	PyModule_AddIntConstant(m, "MNT_MS_FEC_OFFSET", MNT_MS_FEC_OFFSET);
+	PyModule_AddIntConstant(m, "MNT_MS_FEC_ROOTS", MNT_MS_FEC_ROOTS);
 
 	/*
 	 * mount(2) MS_* masks (MNT_MAP_LINUX map)
